// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include "bit_multi_set.h"

namespace bbq {

// ==============================================================	
// ==============================================================	
// ==============================================================	
// ==============================================================	


// ==============================================================
bool bit_multi_set::weighted = true;
// ==============================================================

// ==============================================================
int bit_multi_set::k = 128;
// ==============================================================	

// ==============================================================
int bit_multi_set::k_ = 4;
// ==============================================================	

// ==============================================================	
bit_multi_set::bit_multi_set()
// ==============================================================	
{

	int i;
	counts = new c_uint[k];
	for (i=0; i<k; i++)
		counts[i] = 0;
	weights = new float[k];
	for (i=0; i<k; i++)
		weights[i] = 0.;
	
}

// ==============================================================	
bit_multi_set::bit_multi_set(const bit_multi_set& vb)
// ==============================================================	
{
	int i;
	counts = new c_uint[k];
	for (i=0; i<k; i++)
		counts[i] = vb.counts[i];
}

// ==============================================================	
bit_multi_set& bit_multi_set::operator=(const bit_multi_set& vb)
// ==============================================================	
{
	int i;
	//std::cout<<"+";
	//int k_ = (k>>5)+1;
	for (i=0; i<k; i++)
		counts[i] = vb.counts[i];
	return *this;
}

// ==============================================================	
bool bit_multi_set::subseteq(const bit_multi_set& b) const
// ==============================================================	
{
	
	for (int i=0; i<k; i++)
		if (counts[i]>b.counts[i])
			return false;
	return true;
}

// ==============================================================	
bool bit_multi_set::subseteq(const bit_multi_set& b, const unsigned int& J) const
// ==============================================================	
{
	
	for (int i=0; i<J; i++)
		if (counts[i]>b.counts[i])
			return false;
	return true;
}

// ==============================================================	
bool  bit_multi_set::operator<(const bit_multi_set& b) const
// ==============================================================
{
	for (int i=k; i>0; i--)
		if (counts[i]>0 && b.counts[i]==0)
			return false;
	return true;
}

// ==============================================================	
bool  bit_multi_set::operator==(const bit_multi_set& b) const
// ==============================================================
{
	for (int i=0; i<k; i++)
		if (counts[i]==0 && b.counts[i]!=0 || counts[i]!=0 && b.counts[i]==0)
			return false;
	return true;
}


// ==============================================================	
int bit_multi_set::cardinality()
// ==============================================================	
{
	int card = 0;
	for (int i=0; i<k; i++)
	{
		card += counts[i];
	}
	return card;
}

// ==============================================================	
float bit_multi_set::weight()
// ==============================================================	
{
	return (float)cardinality();
}

// ==============================================================	
void bit_multi_set::intersect(const bit_multi_set& b)
// ==============================================================	
{

	int i;
	//int k_ = (k>>5)+1;
	for (i=0; i<k; i++)
		if (b.counts[i]<counts[i])
			counts[i] = b.counts[i];
}

// ==============================================================	
void bit_multi_set::set(const int& j, const float& w)
// ==============================================================	
{
	counts[j]++;
}

// ==============================================================	
void bit_multi_set::reset(const int& j, const float& w)
// ==============================================================	
{
}

// ==============================================================	
bool bit_multi_set::contains(const int& j)
// ==============================================================	
{
	return counts[j]!=0;
}

// ==============================================================	
void bit_multi_set::remove(const int& j, const float& w)
// ==============================================================	
{
	// this method is useful for sets, not for multisets. See decrease().
}

// ==============================================================	
void bit_multi_set::decrease(const int& j, const float& w)
// ==============================================================	
{
	if (counts[j]>0)
		counts[j]--;
}

// ==============================================================	
std::ostream& operator<<(std::ostream& o,const bit_multi_set& I)
// ==============================================================	
{
	o<<"{ ";
	int j=0;
	for (int i=0; i<bit_multi_set::k; i++)
	{
		o<<(int)I.counts[i]<<" ";
	}
	o<<"}";
	return o;
}


} // end namespace bbq
