// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include "bitsets.h"

namespace bbq {
	
// ==============================================================
int uint_bitset::k = 32;
// ==============================================================	

// ==============================================================
int uint_bitset::cardinality()
// ==============================================================	
{
	int card = 0;
	unsigned int intersection = bits;
	for (; intersection>0; intersection /= 2)
		if (intersection%2==1)
			card++;
	return card;
}

// ==============================================================	
bool uint_bitset::subseteq(const uint_bitset& b)
// ==============================================================	
{
	unsigned int aa =bits;
	unsigned int bb = b.bits;
	
	return (((aa ^ bb) & aa) == 0);
	
}

// ==============================================================	
void uint_bitset::intersect(const uint_bitset& b)
// ==============================================================	
{
	bits &= b.bits;
}


// ==============================================================	
void uint_bitset::set(const int& j)
// ==============================================================	
{
	bits |= 1<<j;
}

// ==============================================================	
void uint_bitset::remove(const int& j)
// ==============================================================	
{
	bits &= UINT_MAX-1<<j;
}

// ==============================================================	
std::ostream& operator<<(std::ostream& o,const uint_bitset& I)
// ==============================================================	
{
	o<<"{";
	unsigned int intersection = I.bits;
	for (int j=0; intersection>0; intersection /= 2, j++)
		if (intersection%2==1)
			o<<j<<", ";
	o<<"}"<<std::endl;
	return o;
}


// ==============================================================	
// ==============================================================	
// ==============================================================	
// ==============================================================	


// ==============================================================
int uint64_bitset::k = 64;
// ==============================================================	

// ==============================================================	
bool uint64_bitset::subseteq(const uint64_bitset& b)
// ==============================================================	
{
	unsigned int aa =bits;
	unsigned int bb = b.bits;
	
	return (((aa ^ bb) & aa) == 0);
	
}


// ==============================================================	
bool uint64_bitset::operator==(const uint64_bitset& vb) const
// ==============================================================	
{
	if (bits!=vb.bits)
		return false;
	return true;
}

// ==============================================================	
bool uint64_bitset::operator<(const uint64_bitset& vb) const
// ==============================================================	
{
	if (bits>vb.bits)
		return false;
	return true;
}

// ==============================================================	
int uint64_bitset::cardinality()
// ==============================================================	
{
	int card = 0;
	unsigned int intersection = bits;
	for (; intersection>0; intersection /= 2)
		if (intersection%2==1)
			card++;
	return card;
}

// ==============================================================	
void uint64_bitset::intersect(const uint64_bitset& b)
// ==============================================================	
{
	bits &= b.bits;
}

// ==============================================================	
void uint64_bitset::set(const int& j)
// ==============================================================	
{
	bits |= 1<<j;
}

// ==============================================================	
void uint64_bitset::remove(const int& j)
// ==============================================================	
{
	bits &= UINT_MAX-1<<j;
}

// ==============================================================	
std::ostream& operator<<(std::ostream& o,const uint64_bitset& I)
// ==============================================================	
{
	o<<"{";
	unsigned int intersection = I.bits;
	for (int j=0; intersection>0; intersection /= 2, j++)
		if (intersection%2==1)
			o<<j<<", ";
	o<<"}"<<std::endl;
	return o;
}


// ==============================================================	
// ==============================================================	
// ==============================================================	
// ==============================================================	


// ==============================================================
int variable_bitset::k = 128;
// ==============================================================	


// ==============================================================	
variable_bitset::variable_bitset()
// ==============================================================	
{

	int k_ = (k>>5)+1;
	bits = new unsigned int[k_];
	for (int i=0; i<k_; i++)
		bits[i] = 0;
	
}

// ==============================================================	
variable_bitset::variable_bitset(const variable_bitset& vb)
// ==============================================================	
{
	//std::cout<<"!";
	int k_ = (k>>5)+1;
	bits = new unsigned int[k_];
	for (int i=0; i<k_; i++)
		bits[i] = vb.bits[i];
}

// ==============================================================	
variable_bitset& variable_bitset::operator=(const variable_bitset& vb)
// ==============================================================	
{
	//std::cout<<"+";
	int k_ = (k>>5)+1;
	for (int i=0; i<k_; i++)
		bits[i] = vb.bits[i];
	return *this;
}

// ==============================================================	
bool variable_bitset::operator==(const variable_bitset& vb)
// ==============================================================	
{
	int i;
	int k_ = (k>>5)+1;	
	for (i=k_; i>0; i--)
		if (bits[i]!=vb.bits[i])
			return false;
	return true;
}

// ==============================================================	
bool variable_bitset::operator<(const variable_bitset& vb) 
// ==============================================================	
{
	int i;
	int k_ = (k>>5)+1;	
	for (i=k_; i>0; i--)
		if (bits[i]>vb.bits[i])
			return false;
	return true;
}

// ==============================================================	
bool variable_bitset::subseteq(const variable_bitset& b) const
// ==============================================================	
{
	
	int k_ = (k>>5)+1;
	bool sse = true;
	for (int i=0; i<k_; i++)
	{
		unsigned int aa = bits[i];
		unsigned int bb = b.bits[i];
		sse = sse && (((aa ^ bb) & aa) == 0);
	}
	return sse;
	
}

// ==============================================================	
int variable_bitset::cardinality()
// ==============================================================	
{
	int card = 0;
	int k_ = (k>>5)+1;
	for (int i=0; i<k_; i++)
	{
		unsigned int intersection = bits[i];
		for (; intersection>0; intersection /= 2)
			if (intersection%2==1)
				card++;
	}
	return card;
}

// ==============================================================	
void variable_bitset::intersect(const variable_bitset& b)
// ==============================================================	
{
	
	int k_ = (k>>5)+1;
	for (int i=0; i<k_; i++)
		bits[i] &= b.bits[i];
}

// ==============================================================	
void variable_bitset::set(const int& j)
// ==============================================================	
{
	bits[j>>5] |= 1<<(j%32);
}

// ==============================================================	
void variable_bitset::remove(const int& j)
// ==============================================================	
{
	bits[j>>5] &= UINT_MAX-1<<(j%32);
}

// ==============================================================	
std::ostream& operator<<(std::ostream& o,const variable_bitset& I)
// ==============================================================	
{
}


} // end namespace bbq
