// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#ifndef CLIQUE_H
#define CLIQUE_H

#include <map>

namespace bbq {

class clique_member
{
public:

	unsigned int val;
	int i;
	int nu;

	bool operator<(const clique_member m) const;
	bool operator==(const clique_member m) const;
	
	clique_member(int i_, int nu_)
		{
			i = i_; nu = nu_;
		}
	
	clique_member()
		{
		}
};

class clique
{

	
public:

	/// The members of the clique should be read
	/// { (0,members[0]), ..., (K-1,members[K-1]) }.
	int* members;

	
	short int color;
	

	/// since during execution, all cliques have same size K, we use a
	/// static variable for storing the cliques' size, which saves us a
	/// whole lot of memory.
	static int K;

	clique();
	
	clique(int*);

	clique(int*, short int);

	clique(const clique& q);
	
	~clique();
	
	clique& operator=(const clique& q);

	bool operator==(const clique& q) const;

	bool operator<(const clique& q) const;
	
};

class clique_features
{
public:

	clique_features(clique*,int);

	clique_features() {};
	
	clique* best_path_pred;
	int best_path_length;	
	
};


typedef std::map<clique,clique_features> clique_map;

} // end namespace bbq

#endif
