// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include <ctype.h>
#include <iostream>
#include <fstream>
#include <list>
#include <stdio.h>
#include <map>
#include <string>
#include <set>


#define CHUNK_SIZE 4096

	
bool params_good;
int num_of_opts;

char** s;
char** s_names;
int* n;
int* m;
int L;

// number of fragments
int k;

// number of groups
int kappa;

int K;

std::set<std::string> groups;
std::map< std::string,std::list<int> > group_members;

void fill_reverse_segments();

/// consider reverse complemented segments also?
bool reverse;



/// read over all useless characters such as spaces, tabs and line
/// breaks
void skip(std::ifstream& ifs)
{
	if (!ifs.good())
		return;
	char ch = 0;
	ifs.get(ch);
	while (ifs.good() && ch<=32)
		ifs.get(ch);
	if (ifs.good())
		ifs.unget();
}

bool read_parameters(int argc, char** argv)
{
	
	params_good = true;
	if (argc<1)
	{
		params_good = false;
		return false;
	}
	
	reverse = false;
	num_of_opts = 0;
	int i;
	for (i=1; i<argc && argv[i][0]=='-'; i++)
	{
		num_of_opts++;
		switch (argv[i][1])
		{
			case 'r' : {
				reverse = true;
				break;
			}
			default: {
				std::cout<<"ignoring unknown option "<<argv[i][1]<<std::endl;
			}
		}
	}
	return true;      
}


void read_data(int argc, char** argv)
{
	
	int j,l;
	params_good = true;

	
	// I. READ FRAGMENT SEQUENCES s_1,...,s_k from the fragment-file
	std::list<char*> s_list, s_name_list;
	std::list<int> n_list, m_list;
	std::ifstream frag_file(argv[num_of_opts+1]);
	skip(frag_file);	
	while (frag_file.good())
	{
		char read_buf[1024];
		// read fragment name
		read_buf[0]=0;
		read_buf[1023] = 0;
		int name_len;
		frag_file.getline(read_buf,1024,' ');
		name_len = strlen(read_buf);
		char* curr_name = new char[name_len+1];
		strcpy(curr_name,read_buf);
		s_name_list.push_back(curr_name);
		
		skip(frag_file);
		
		// read fragment sequence, currently delimited to 1024 characters
		read_buf[0]=0;
		read_buf[1023] = 0;
		int s_len;
		frag_file.getline(read_buf,1024,' ');
		s_len = strlen(read_buf);
		char* curr_s = new char[s_len+1];
		strcpy(curr_s,read_buf);
		s_list.push_back(curr_s);
		n_list.push_back(s_len);
		skip(frag_file);
		
		// read maximum edit distance for the fragment
		read_buf[0]=0;
		read_buf[1023] = 0;
		int d_len;
		frag_file.getline(read_buf,1024);
		m_list.push_back(atoi(read_buf));
		skip(frag_file);		
	}

	k = s_list.size();
	if (reverse)
		k *= 2;
	s = new char*[k];
	m = new int[k];
	n = new int[k];
	s_names = new char*[k];
	
	std::list<char*>::iterator s_it = s_list.begin();
	std::list<char*>::iterator names_it = s_name_list.begin();
	std::list<int>::iterator n_it = n_list.begin();
	std::list<int>::iterator m_it = m_list.begin();
	for (j=0; j<k; j++, ++s_it, ++n_it, ++names_it, ++m_it)
	{
		n[j] = *n_it;
		m[j] = *m_it;
		s[j] = *s_it;
		s_names[j] = *names_it;
		int iota;
		char* s_j = s[j];
		int n_j = n[j];
		for (iota=0; iota<n_j; iota++)
			s_j[iota] = toupper(s_j[iota]);
	}
	
	if (reverse)
		fill_reverse_segments();
	
}

void fill_reverse_segments()
{

	if (!reverse)
		return;
	
	int i;
	
	char* compl_table = new char[256];

	for(i=0; i<256; i++) compl_table[i]=(int)'-';
	
	compl_table[(int)'A']='T';
	compl_table[(int)'C']='G';
	compl_table[(int)'G']='C';
	compl_table[(int)'T']='A';
	compl_table[(int)'R']='Y';
	compl_table[(int)'Y']='R';
	compl_table[(int)'M']='K';
	compl_table[(int)'K']='M';
	compl_table[(int)'S']='S';
	compl_table[(int)'W']='W';
	compl_table[(int)'H']='D';
	compl_table[(int)'B']='V';
	compl_table[(int)'V']='B';
	compl_table[(int)'D']='H';
	compl_table[(int)'N']='N';
	compl_table[(int)'a']='t';
	compl_table[(int)'c']='g';
	compl_table[(int)'g']='c';
	compl_table[(int)'t']='a';
	compl_table[(int)'r']='y';
	compl_table[(int)'y']='r';
	compl_table[(int)'m']='k';
	compl_table[(int)'k']='m';
	compl_table[(int)'s']='s';
	compl_table[(int)'w']='w';
	compl_table[(int)'h']='d';
	compl_table[(int)'b']='v';
	compl_table[(int)'v']='b';
	compl_table[(int)'d']='h';
	compl_table[(int)'n']='n';
	
	int j, iota;
	// k is divisible by two, since it has been mult by two due to
	// reverse option before.
	int k_2 = k/2;
	for (j=0; j<k_2; j++)
	{
		int n_j = n[j];
		s[k_2+j] = new char[n[j]+1];
		n[k_2+j] = n[j];
		s_names[k_2+j] = new char[strlen(s_names[j])+5];
		strcpy(s_names[k_2+j],s_names[j]);
		strcat(s_names[k_2+j],"-rev");
		for (iota=0; iota<n_j; iota++)
			s[k_2+j][iota] = compl_table[s[j][n_j-iota-1]];
		s[k_2+j][n_j] = (char)0;
	}
}


void free_all()
{
	int i,j;
	for (j=0; j<k; j++)
		delete[] s[j];
	delete[] s;
	s = NULL;
	for (j=0; j<k; j++)
		delete[] s_names[j];
	delete[] s_names;
	s_names = NULL;
	delete[] n;
	n = NULL;
	m = NULL;
	
}

bool matches(int i, int j)
{
	int mu,nu,mismatches, min_mismatches;
	min_mismatches = INT_MAX;
	for (mu=0; mu<n[i]; mu++)
	{
		mismatches = 0;
		// does s[j] occur with at most m[j] mismatches in s[i] at pos. mu?
		for (nu=0; nu<n[j]; nu++)
		{
			if (mu+nu>=n[i] || s[i][mu+nu]!=s[j][nu])
				mismatches++;
		}
		if (mismatches<min_mismatches)
			min_mismatches = mismatches;
		if (min_mismatches<=m[j])
			// s[j] occurs w/ at most m[j] mismatches in s[i]. Hence,
			// each occurence of s[i] induces an occurence of s[j], so
			// they should be grouped together.
			return true;
	}
	return false;
}

void process_data()
{
	int j;
	bool matched;
	
	kappa = 1;
	std::string group_name;
	group_name = std::string("G001");
	groups.insert(group_name);
	group_members[group_name].push_back(0);
	if (reverse)
		group_members[group_name].push_back(k/2);
	int last_k;
	if (reverse)
		last_k = k/2;
	else
		last_k = k;
		
	for (j=1; j<last_k; j++)
	{
		matched = false;
		std::set< std::string >::iterator g_it = groups.begin();
		std::set< std::string >::iterator g_end = groups.end();
		for (; g_it != g_end && !matched; ++g_it)
		{
			matched = false;
			std::list<int>::iterator l_it = group_members[*g_it].begin();
			std::list<int>::iterator l_end = group_members[*g_it].end();
			for (; l_it!=l_end && !matched; ++l_it)
			{
				if (matches(j,*l_it) || matches(*l_it,j))
				{
					matched = true;
				}
				if (reverse)
					if (matches(j+k/2,*l_it) || matches(*l_it,j+k/2))
					{
						matched = true;
					}

			}
			if (matched)
			{
				group_members[*g_it].push_back(j);
				if (reverse)
					group_members[*g_it].push_back(j+k/2);
			} 
		}
		if (!matched)
		{
			//create a new group
			kappa++;
			if (kappa>=1000)
			{
				std::cout<<"sorry, too many groups to continue; number of groups is limited to 1000.\n";
			}
			char buf[5];
			buf[0] = 'G';
			int k_ = kappa;
			buf[4] = 0;
			buf[3] = 48+(k_%10); 
			k_ /= 10;
			buf[2] = 48+(k_%10);
			k_ /= 10;
			buf[1] = 48+(k_%10);
			group_name = std::string(buf);
			groups.insert(group_name);
			group_members[group_name].push_back(j);
			if (reverse)
				group_members[group_name].push_back(j+k/2);
			}
	}
}

void write_output()
{
	int i;
	std::cerr<<groups.size()<<" groups created from "<<k<<" fragments\n";
		std::set< std::string >::iterator g_it = groups.begin();
		std::set< std::string >::iterator g_end = groups.end();
		for (; g_it != g_end; ++g_it)
		{
			std::list<int>::iterator l_it = group_members[*g_it].begin();
			std::list<int>::iterator l_end = group_members[*g_it].end();
			for (; l_it!=l_end; ++l_it)
			{
				if (!reverse || *l_it<k/2)
					std::cout<<*g_it<<s_names[*l_it]<<" "<<s[*l_it]<<" "<<m[*l_it]<<"\n";
			}
		}
}

void print_help()
{
	std::cerr<<"bbqcluster [-r] <filename>\n <filename> is the motif file with the "
				<<"fragments to be clustered;\n if -r is applied, reverse segments are "
				<<"taken into account for clustering.\n";
}

int main(int argc, char** argv)
{
	
	s = NULL;
	s_names = NULL;
	n = NULL;
	m = NULL;
	if (argc<=1)
	{
		print_help();
		return 0;
	}
	
	std::cerr<<"\nclustering fragments....\n\n";

	read_parameters(argc,argv);
	if (!params_good)
	{
		print_help();
		return 0;
	}
	read_data(argc,argv);
	if (!params_good)
	{
		print_help();		
		return 0;
	}
	process_data();
	write_output();
	free_all();

	return 0;
	
}


