// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#ifndef FOOTPRINT_DETECTOR_H
#define FOOTPRINT_DETECTOR_H

#include <set>
#include "bitsets.h"


namespace bbq {
	
//typedef uint_bitset bit_set;
//typedef uint64_bitset bit_set;
typedef variable_bitset bit_set;

/**
	The class footprint_detector exists for rather historical
	reasons. It only supports unweighted barbeques and in
	implementation of the branch-and-bound version of (A1), without
	grouping or reversing elements. However, it demonstrates with how
	little programming effort (hardly more than 300 lines of code) the
	best-barbeque technique can be implemented.

	This simplicity unfortunately gets lost in the "real"
	implementation, so please enjoy this piece of code :)
 */
class footprint_detector
{
	
public:


private:
	
	/// number of fragment sequences s_1,...,s_k
	int k;

	/// number of genome (excerpts) T_1,...,T_K
	int K;

	/// fragment sequences s_1,...,s_k
	char** s;
	
	/// genome sequences T_1,...,T_K
	char** T;

	/// n_j contains the length of s_j
	int* n;

	/// N_i contains the length of T_i
	int* N;

	int *** p;
	int ** lambda;
	int * Lambda;
	bit_set ** C;
	int ** P;
	
	int L;

	/// This variable contains the result of the optimization
	/// procedure.
	int overall_best_length;

	int* best_nu;
	
	bit_set best_intersection;

	
public:

	footprint_detector();
	
	footprint_detector(int,int,char**,char**,int*,int*);

	~footprint_detector();

	void init_arrangement_cells();

	void compute_intersections();

	void get_occurences(int i, int j);
	
	void run(int);

	int get_result();

	int print_result();

	int intersection_cardinality(unsigned int);

	bool increase(int*,int);
	
};

} // end namespace bbq


#endif
