// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include "weighted_bitsets.h"

namespace bbq {

// ==============================================================	
// ==============================================================	
// ==============================================================	
// ==============================================================	


// ==============================================================
bool weighted_variable_bitset::weighted = true;
// ==============================================================

// ==============================================================
int weighted_variable_bitset::k = 128;
// ==============================================================	

// ==============================================================
int weighted_variable_bitset::k_ = 4;
// ==============================================================	


// ==============================================================	
weighted_variable_bitset::~weighted_variable_bitset()
// ==============================================================	
{
	if (bits!=NULL)
	{
		delete[] bits;
		bits = NULL;
	}
	if (weights!=NULL)
	{
		delete[] weights;
		weights = NULL;
	}
}

// ==============================================================	
weighted_variable_bitset::weighted_variable_bitset()
// ==============================================================	
{

	int i;
	bits = new unsigned int[k_];
	for (i=0; i<k_; i++)
		bits[i] = 0;
	weights = new float[k];
	for (i=0; i<k; i++)
		weights[i] = 0.;
	
}

// ==============================================================	
weighted_variable_bitset::weighted_variable_bitset(const weighted_variable_bitset& vb)
// ==============================================================	
{
	int i;
	//std::cout<<"!";
	//int k_ = (k>>5)+1;
	bits = new unsigned int[k_];
	for (i=0; i<k_; i++)
		bits[i] = vb.bits[i];
	weights = new float[k];
	for (i=0; i<k; i++)
		weights[i] = vb.weights[i];
}

// ==============================================================	
weighted_variable_bitset& weighted_variable_bitset::operator=(const weighted_variable_bitset& vb)
// ==============================================================	
{
	int i;
	//std::cout<<"+";
	//int k_ = (k>>5)+1;
	if (bits!=NULL)
		delete[] bits;
	bits = new unsigned int[k_];
	if (weights!=NULL)
		delete[] weights;
	weights = new float[k];
	
	for (i=0; i<k_; i++)
		bits[i] = vb.bits[i];
	for (i=0; i<k; i++)
		weights[i] = vb.weights[i];
	return *this;
}

// ==============================================================	
bool weighted_variable_bitset::operator==(const weighted_variable_bitset& vb) const
// ==============================================================	
{
	int i;
	for (i=k_-1; i>=0; i--)
		if (bits[i]!=vb.bits[i])
			return false;
	return true;
}

// ==============================================================	
bool weighted_variable_bitset::operator<(const weighted_variable_bitset& vb) const
// ==============================================================	
{
	int i;
	for (i=k_-1; i>=0; i--)
		if (bits[i]>vb.bits[i])
			return false;
		else if (bits[i]<vb.bits[i])
			return true;
	return false;
}

// ==============================================================	
bool weighted_variable_bitset::subseteq(const weighted_variable_bitset& b) const
// ==============================================================	
{
	
	//int k_ = (k>>5)+1;
	for (int i=0; i<k_; i++)
	{
		unsigned int aa = bits[i];
		unsigned int bb = b.bits[i];
		if (((aa ^ bb) & aa) != 0)
			return false;
	}
	return true;
	
}

// ==============================================================	
bool weighted_variable_bitset::subseteq(const weighted_variable_bitset& b, const unsigned int& J) const
// ==============================================================	
{
	
	int k_ = (J>>5)+1;
	for (int i=0; i<k_; i++)
	{
		unsigned int aa = bits[i];
		unsigned int bb = b.bits[i];
		if (((aa ^ bb) & aa) != 0)
			return false;
	}
	return true;
	
}

// ==============================================================	
int weighted_variable_bitset::cardinality()
// ==============================================================	
{
	int card = 0;
	//int k_ = (k>>5)+1;
	for (int i=0; i<k_; i++)
	{
		unsigned int intersection = bits[i];
		for (; intersection>0; intersection /= 2)
			if (intersection%2==1)
				card++;
	}
	return card;
}

// ==============================================================	
float weighted_variable_bitset::weight()
// ==============================================================	
{
	if (weighted==false)
		return (float)cardinality();
	
	float wght = 0;
	int j=0;
	//int k_ = (k>>5)+1;
	for (int i=0; i<k_; i++)
	{
		unsigned int intersection = bits[i];
		for (int l=0; l<32; intersection /= 2, j++, l++)
			if (intersection%2==1)
				wght += weights[j];
	}
	return wght;
}

// ==============================================================	
void weighted_variable_bitset::intersect(const weighted_variable_bitset& b)
// ==============================================================	
{

	int i;
	//int k_ = (k>>5)+1;
	for (i=0; i<k_; i++)
		bits[i] &= b.bits[i];
	for (i=0; i<k; i++)
		weights[i] += b.weights[i];
}

// ==============================================================	
void weighted_variable_bitset::set(const int& j, const float& w)
// ==============================================================	
{
	bits[j>>5] |= 1<<(j%32);
	weights[j] = w;
}

// ==============================================================	
void weighted_variable_bitset::reset(const int& j, const float& w)
// ==============================================================	
{
	bits[j>>5] |= 1<<(j%32);
	weights[j] = w;
}

// ==============================================================	
bool weighted_variable_bitset::contains(const int& j)
// ==============================================================	
{
	unsigned int res = (bits[j>>5]) & (1<<(j%32));
	if (res==0)
		return false;
	return true;
}

// ==============================================================	
void weighted_variable_bitset::remove(const int& j, const float& w)
// ==============================================================	
{
	bits[j>>5] &= UINT_MAX-(1<<(j%32));
	weights[j] = 0.;
}

// ==============================================================	
void weighted_variable_bitset::decrease(const int& j, const float& w)
// ==============================================================	
{
	// nothing to be done, this method is needed for handling multisets only.
}

// ==============================================================	
std::ostream& operator<<(std::ostream& o,const weighted_variable_bitset& I)
// ==============================================================	
{
	o<<"{";
	int j=0;
	//int k_ = (I.k>>5)+1;
	for (int i=0; i<weighted_variable_bitset::k_; i++)
	{
		unsigned int intersection = I.bits[i];
		for (int l=0; l<32; intersection /= 2, j++, l++)
			if (intersection%2==1)
			{
				o<<"s_"<<j+1;
				if (!(i==weighted_variable_bitset::k_-1 && intersection==1))
					o<<", ";
			}
	}
	o<<"}";
	return o;
}

} // end namespace bbq

